
# This file is automatically generated by setup.py, So don't edit it. :)
VERSION_INFO = {
    'final': True,
    'version': '1.8.2',
    'branch_nick': '(no',
    'revision_id': '218e740e02479a88639f61439e5655e73899dad8',
    'revno': 4480
}
